/*
 * NUCLEO F411/F401 with HC-SR04 distance sensor
 * gbm 12'2015
 */
#include "board.h"
#include "timing.h"

#define	TRIG_BIT	10
#define ECHO_BIT	9

#define S_CKFREQ	100000	// time unit freq. Hz
#define S_MFREQ		4	// measure freq. Hz
#if S_CKFREQ / S_MFREQ > 1 << 16
#error Bad timing!
#endif
#define TRIG_WIDTH	(S_CKFREQ / 100000 + 1)
#define RST_START	(S_CKFREQ / S_MFREQ - 3 * S_CKFREQ / 100000 - 1)
#define MAX_V_PING	(1000 * S_CKFREQ / 1000000)

int main(void)
{
	// enable peripherals
	RCC->AHB1ENR = RCC_AHB1ENR_GPIOAEN | RCC_AHB1ENR_DMA1EN;
	RCC->APB1ENR = RCC_APB1ENR_USART2EN;
	RCC->APB2ENR = RCC_APB2ENR_TIM1EN;
	// port setup
	GPIOA->OTYPER = 1u << ECHO_BIT;	// Echo OD for reset
	GPIOA->AFR[0] = BF4(RX_BIT, 7) | BF4(TX_BIT, 7); // USART2
	GPIOA->AFR[1] = BF4(TRIG_BIT, 1) | BF4(ECHO_BIT, 1); // TIM1CH3,2
	GPIOA->MODER = GPIOA_MODER_SWD
		| BF2(TRIG_BIT, GPIO_MODER_AF) | BF2(ECHO_BIT, GPIO_MODER_AF)	// TIM1CH3,2 - trig, echo
		| BF2(TX_BIT, GPIO_MODER_AF) | BF2(RX_BIT, GPIO_MODER_AF);	// UART2 pins as AF
	// USART2 setup - PC connection
	USART2->BRR = (APB1_FREQ + BAUD_RATE / 2) / BAUD_RATE;
	USART2->CR3 = USART_CR3_DMAT;	// enable TX DMA
	USART2->CR1 = USART_CR1_TE | USART_CR1_UE;	// enable
	// TIM1 - distance sensor; CH1 - echo, CH2 - echo reset, CH3 - trig
	TIM1->PSC = APB2_TFREQ / S_CKFREQ - 1;	// 10 us
	TIM1->ARR = S_CKFREQ / S_MFREQ - 1;	// 2 Hz
	TIM1->CCR3 = TRIG_WIDTH;	// trigger pulse >= 10 us
	TIM1->CCR2 = RST_START;	// reset pulse - 20 us
	TIM1->CCMR1 = TIM_CCMR1_OC2M_PWM1 | TIM_CCMR1_CC1S_1;	// PWM mode 1, capture from TI2
	TIM1->CCMR2 = TIM_CCMR2_OC3M_PWM1;	// PWM mode 1
	TIM1->CCER = TIM_CCER_CC3E | TIM_CCER_CC2E | TIM_CCER_CC1NP | TIM_CCER_CC1P | TIM_CCER_CC1E;	// enable CH3&2 output, CH1 capture both edges
	TIM1->BDTR = TIM_BDTR_MOE;	// enable PWM outputs
	TIM1->DIER = TIM_DIER_CC1IE;	// enable capture 1 interrupt
	TIM1->CR1 = TIM_CR1_CEN;	// enable
	// interrupts and sleep
	NVIC_EnableIRQ(TIM1_CC_IRQn);
	SCB->SCR = SCB_SCR_SLEEPONEXIT_Msk;	// sleep while not in handler
	__WFI();
}

void TIM1_CC_IRQHandler(void)
{
	static char diststr[] = "   \r\n";
	static uint16_t	ping, echo;
	
	uint32_t ncap = TIM1->CCR1;
	
	if (TIM1->SR & TIM_SR_UIF)
	{
		TIM1->SR = ~TIM_SR_UIF;
		if (ping > TRIG_WIDTH && ping < MAX_V_PING && echo > ping)
		{
			uint32_t dist_cm = echo >= RST_START ? 999
				: (echo - ping) * (1000000 / S_CKFREQ) / 58;
			// convert distance to string
			diststr[2] = dist_cm % 10 + '0';
			dist_cm /= 10;
			diststr[1] = dist_cm % 10 + '0';
			dist_cm /= 10;
			diststr[0] = dist_cm % 10 + '0';
			// start DMA xfer
			DMA1_Stream6->CR = 0;	// disable
			DMA1->HIFCR = DMA_HIFCR_CALLF6;	// clear stream flags
			DMA1_Stream6->PAR = (uint32_t)&USART2->DR;
			DMA1_Stream6->M0AR = (uint32_t)diststr;
			DMA1_Stream6->NDTR = sizeof(diststr) - 1;
			// USART2_TX, increment memory adress, mem->periph, enable
			DMA1_Stream6->CR = DMA_SxCR_CHSELV(4)
				| DMA_SxCR_MINC | DMA_SxCR_DIR_M2P | DMA_SxCR_EN;
		}
		ping = ncap;
	}
	else
		echo = ncap;
}
